from collections.abc import Mapping
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from typing import Any, Optional, Union

import numpy as np
from tqdm import tqdm

from skllm.llm.base import (
    BaseClassifierMixin,
    BaseEmbeddingMixin,
    BaseTextCompletionMixin,
)
from skllm.utils import extract_json_key
from skollama.llm.ollama.completion import get_chat_completion
from skollama.llm.ollama.embedding import get_embedding


def construct_message(role: str, content: str) -> dict:
    """Constructs a message for the OpenAI API.

    Parameters
    ----------
    role : str
        The role of the message. Must be one of "system", "user", or "assistant".
    content : str
        The content of the message.

    Returns
    -------
    message : dict
    """
    if role not in ("system", "user", "assistant"):
        raise ValueError("Invalid role")
    return {"role": role, "content": content}


class OllamaCompletionMixin(BaseTextCompletionMixin):
    """Mixin for handling chat completions with the Ollama server.

    This mixin provides methods to interact with the Ollama server for obtaining chat completions.
    It defines the following functionality:

    Methods
    -------
    _get_chat_completion(model, messages, system_message=None, **kwargs)
        Gets a chat completion from the Ollama server based on the provided messages and model.

    _convert_completion_to_str(completion)
        Converts the completion object returned by the Ollama server into a string.
    """

    def _get_chat_completion(
        self,
        model: str,
        messages: Union[str, list[dict[str, str]]],
        system_message: Optional[str] = None,
        **kwargs: Any,
    ):
        """Gets a chat completion from the Ollama server.

        Parameters
        ----------
        model : str
            The model to use.
        messages : Union[str, List[Dict[str, str]]]
            input messages to use.
        system_message : Optional[str]
            A system message to use.
        **kwargs : Any
            placeholder.

        Returns
        -------
        completion : dict
        """
        msgs = []
        if system_message is not None:
            msgs.append(construct_message("system", system_message))
        if isinstance(messages, str):
            msgs.append(construct_message("user", messages))
        else:
            for message in messages:
                msgs.append(construct_message(message["role"], message["content"]))
        completion = get_chat_completion(
            msgs,
            model,
            self.host,
            self.options,
            self.format,
        )
        return completion

    def _convert_completion_to_str(self, completion: Mapping[str, Any]):
        if hasattr(completion, "__getitem__"):
            return str(completion["message"]["content"])
        return str(completion.message.content)


class OllamaClassifierMixin(OllamaCompletionMixin, BaseClassifierMixin):
    """Mixin for Ollama-based classification tasks.

    This mixin extends the OllamaCompletionMixin to provide functionality
    specific to classification tasks, leveraging the Ollama model's ability
    to classify text based on provided inputs and completions.

    Methods
    -------
    _extract_out_label(completion, **kwargs)
        Extracts the classification label from the Ollama model's completion.
    """

    def _extract_out_label(self, completion: Mapping[str, Any], **kwargs) -> Any:
        """Extracts the label from a completion.

        Parameters
        ----------
        label : Mapping[str, Any]
            The label to extract.

        Returns
        -------
        label : str
        """
        if self._base_model:
            key = list(self._base_model.__fields__.keys())[0]
        else:
            key = "label"
        try:
            if hasattr(completion, "__getitem__"):
                label = extract_json_key(completion["message"]["content"], key)
            else:
                label = extract_json_key(completion.message.content, "label")
        except Exception as e:
            print(completion)
            print(f"Could not extract the label from the completion: {str(e)}")
            label = ""
        return label


class OllamaEmbeddingMixin(BaseEmbeddingMixin):
    """Mixin for Ollama-based embedding tasks.

    This mixin extends the BaseEmbeddingMixin to provide a multi-threaded approach
    for getting embeddings.
    It defines the following methods:

    Methods
    -------
    _get_embeddings(self, text: np.ndarray)
        Queries the locally running embedding server with the provided texts.
    """

    def _get_embeddings(self, text: np.ndarray) -> list[list[float]]:
        """Gets embeddings from the OpenAI compatible API.

        Parameters
        ----------
        text : str
            The text to embed.
        model : str
            The model to use.
        batch_size : int, optional
            The batch size to use. Defaults to 1.

        Returns
        -------
        embedding : List[List[float]]
        """
        embeddings = []
        print(
            "Batch size:", self.batch_size
        )  # does not work yet, needs refactor of probably WAY more things
        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            embs = list(
                tqdm(
                    executor.map(
                        lambda x, y: get_embedding(x, y),
                        text,
                        repeat(self.model, len(text)),
                    ),
                    total=len(text),
                )
            )
        for i in embs:
            embeddings.extend(i)
        # for i in tqdm(range(0, len(text), self.batch_size)):
        # batch = text[i : i + self.batch_size].tolist()
        # embeddings.extend( # technically a single instance, can be multiprocessed to allow for batches
        #     get_embedding(
        #         batch,
        #         self.model,
        #     )
        # )

        return embeddings



# skollama/classification/__init__.py
"""Text classification models using large language models."""

from skollama.models.ollama.classification.zero_shot import (
    ZeroShotOllamaClassifier,
    MultiLabelZeroShotOllamaClassifier,
    CoTOllamaClassifier,
)

from skollama.models.ollama.classification.few_shot import (
    FewShotOllamaClassifier,
    DynamicFewShotOllamaClassifier,
    MultiLabelFewShotOllamaClassifier,
)

# TODO add OllamaTunableClassifier

__all__ = [
    "ZeroShotOllamaClassifier",
    "MultiLabelZeroShotOllamaClassifier",
    "FewShotOllamaClassifier",
    "DynamicFewShotOllamaClassifier",
    "MultiLabelFewShotOllamaClassifier",
    "CoTOllamaClassifier",
]



## Ollama
# TODO add text2text tunable
from skollama.models.ollama.text2text.summarization import OllamaSummarizer
from skollama.models.ollama.text2text.translation import OllamaTranslator

__all__ = ["OllamaSummarizer", "OllamaTranslator"]



from typing import Optional

from skllm.models._base.text2text import BaseSummarizer
from skollama.llm.ollama.mixin import OllamaCompletionMixin


class OllamaSummarizer(
    BaseSummarizer,
    OllamaCompletionMixin,
):
    """Text summarizer using Ollama API compatible models.

    Attributes
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    max_words : Optional[int], optional
        maximum number of words to use, by default 15
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        max_words: int = 15,
        focus: Optional[str] = None,
    ):
        self.model = model
        self.max_words = max_words
        self.focus = focus
        self.host = host
        self.options = options



from pydantic import BaseModel

from skollama.llm.ollama.clients.ollama.completion import (
    get_chat_completion as _ollama_get_chat_completion,
)


def get_chat_completion(
    messages: dict,
    model: str = "llama3",
    host: str = "http://localhost:11434/",
    options: dict = None,
    format: BaseModel = "",
):
    """Gets a chat completion from an ollama server running locally or remote.

    Parameters
    ----------
    messages : dict
        dictionary with the chat history
    model : str, optional
        model to use, by default "llama3"
    host : str, optional
        Ollama host to connect to, by default "http://localhost:11434/"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    """
    return _ollama_get_chat_completion(
        messages=messages, model=model, host=host, options=options, format=format
    )



from typing import Optional

from pydantic import BaseModel

from skllm.models._base.classifier import BaseCoTClassifier as _BaseCoTClassifier
from skllm.models._base.classifier import (
    BaseZeroShotClassifier as _BaseZeroShotClassifier,
)
from skllm.models._base.classifier import MultiLabelMixin as _MultiLabelMixin
from skllm.models._base.classifier import SingleLabelMixin as _SingleLabelMixin
from skollama.llm.ollama.mixin import OllamaClassifierMixin as _OllamaClassifierMixin


class ZeroShotOllamaClassifier(
    _BaseZeroShotClassifier, _OllamaClassifierMixin, _SingleLabelMixin
):
    """Zero-shot text classifier using ollama API-compatible models.

    Parameters
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    prompt_template : Optional[str], optional
        custom prompt template to use, by default None
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        default_label: str = "Random",
        prompt_template: Optional[str] = None,
        structured_output: Optional[BaseModel] = "",
        **kwargs,
    ):
        super().__init__(
            model=model,
            default_label=default_label,
            prompt_template=prompt_template,
            **kwargs,
        )
        self.host = host
        self.options = options
        self._base_model = structured_output
        if structured_output and issubclass(structured_output, BaseModel):
            json_schema = structured_output.model_json_schema()
        else:
            json_schema = ""
        self.format = json_schema


class CoTOllamaClassifier(
    _BaseCoTClassifier, _OllamaClassifierMixin, _SingleLabelMixin
):
    """Chain-of-though text classifier using Ollama API compatible models.

    Attributes
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    max_labels : Optional[int], optional
        maximum labels per sample, by default 5
    prompt_template : Optional[str], optional
        custom prompt template to use, by default None
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        default_label: str = "Random",
        prompt_template: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            model=model,
            default_label=default_label,
            prompt_template=prompt_template,
            **kwargs,
        )
        self.host = host
        self.options = options


class MultiLabelZeroShotOllamaClassifier(
    _BaseZeroShotClassifier, _OllamaClassifierMixin, _MultiLabelMixin
):
    """Multi-label zero-shot text classifier using Ollama API-compatible
    models.

    Attributes
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    max_labels : Optional[int], optional
        maximum labels per sample, by default 5
    prompt_template : Optional[str], optional
        custom prompt template to use, by default None
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        default_label: str = "Random",
        max_labels: Optional[int] = 5,
        prompt_template: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            model=model,
            default_label=default_label,
            max_labels=max_labels,
            prompt_template=prompt_template,
            **kwargs,
        )
        self.host = host
        self.options = options



import warnings

from skllm.models._base.vectorizer import BaseVectorizer as _BaseVectorizer
from skollama.llm.ollama.mixin import OllamaEmbeddingMixin as _OllamaEmbeddingMixin


# TODO refactor to use client instead
class OllamaVectorizer(_BaseVectorizer, _OllamaEmbeddingMixin):
    """An Ollama based embedding model for vectorizing text.

    This vectorizer leverages Ollama served models to create text embeddings.

    Attributes
    ----------
    model : str, optional
        model to use, by default "custom_url::nomic-embed-text"
    batch_size : int, optional
        number of samples per request, by default 1
    url: str, optional
        url to use for the embedding queries, by default localhost

    Warnings
    --------
    OllamaVectorizer is an experimental feature and may undergo significant changes in future releases.
    """

    def __init__(
        self,
        model: str = "custom_url::nomic-embed-text",
        batch_size: int = 1,
        url: str = "http://localhost:11434/api/embeddings",
    ):
        if batch_size > 1:
            warnings.warn("Batch size controlls the number of parallel embedding calls")
        super().__init__(model=model, batch_size=batch_size)
        self.url = url
        self.num_workers = batch_size
        warnings.warn(
            "OllamaVectorizer is an experimental feature and will see some refactoring"
            " in the future."
        )



from skllm.models._base.text2text import BaseTranslator
from skollama.llm.ollama.mixin import OllamaCompletionMixin


class OllamaTranslator(
    BaseTranslator,
    OllamaCompletionMixin,
):
    """Text summarizer using Ollama API compatible models.

    Attributes
    ----------
    model : str, optional
        model to use, by default "gpt-3.5-turbo"
    output_language : str, optional
        language to translate to, by default "English"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    """

    def __init__(
        self,
        model: str = "llama3",
        output_language: str = "English",
        host: str = "http://localhost:11434",
        options: dict = None,
    ):
        self.model = model
        self.output_language = output_language
        self.host = host
        self.options = options



from skllm.llm.gpt.utils import split_to_api_and_model
from skollama.llm.ollama.clients.ollama.embedding import (
    get_embedding as _ollama_get_embedding,
)


def get_embedding(
    text: str,
    model: str = "custom_url::llama3",
):
    """Encodes a string and return the embedding for a string.

    Parameters
    ----------
    text : str
        The string to encode.
    model : str
        model to use, by default "llama3"

    Returns
    -------
    emb : list
        The embedding for the string.
    """
    api, model = split_to_api_and_model(model)
    if api == "gpt4all":
        raise ValueError("GPT4All is not supported for embeddings")
    return _ollama_get_embedding(text, model, api)



from typing import Optional

from pydantic import BaseModel

from skllm.memory.base import IndexConstructor
from skllm.models._base.classifier import (
    BaseDynamicFewShotClassifier,
    BaseFewShotClassifier,
    MultiLabelMixin,
    SingleLabelMixin,
)
from skllm.models._base.vectorizer import BaseVectorizer
from skollama.llm.ollama.mixin import OllamaClassifierMixin
from skollama.models.ollama.vectorization import OllamaVectorizer


class FewShotOllamaClassifier(
    BaseFewShotClassifier, OllamaClassifierMixin, SingleLabelMixin
):
    """Few-shot text classifier using Ollama API-compatible models.

    Attributes
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    prompt_template : Optional[str], optional
        custom prompt template to use, by default None
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        default_label: str = "Random",
        prompt_template: Optional[str] = None,
        structured_output: Optional[BaseModel] = "",
        **kwargs,
    ):
        super().__init__(
            model=model,
            default_label=default_label,
            prompt_template=prompt_template,
            **kwargs,
        )
        self.host = host
        self.options = options
        self._base_model = structured_output
        if structured_output and issubclass(structured_output, BaseModel):
            json_schema = structured_output.model_json_schema()
        else:
            json_schema = ""
        self.format = json_schema


class MultiLabelFewShotOllamaClassifier(
    BaseFewShotClassifier, OllamaClassifierMixin, MultiLabelMixin
):
    """Multi-label few-shot text classifier using Ollama API-compatible models.

    Parameters
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    prompt_template : Optional[str], optional
        custom prompt template to use, by default None
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        default_label: str = "Random",
        max_labels: Optional[int] = 5,
        prompt_template: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            model=model,
            default_label=default_label,
            max_labels=max_labels,
            prompt_template=prompt_template,
            **kwargs,
        )
        self.host = host
        self.options = options


class DynamicFewShotOllamaClassifier(
    BaseDynamicFewShotClassifier, OllamaClassifierMixin, SingleLabelMixin
):
    """Dynamic few-shot text classifier using Ollama API-compatible models. For
    each sample, N closest examples are retrieved from the memory.

    Parameters
    ----------
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    default_label : str, optional
        default label for failed prediction; if "Random" -> selects randomly based on class frequencies, by default "Random"
    prompt_template : Optional[str], optional
        custom prompt template to use, by default None
    n_examples : int, optional
        number of closest examples per class to be retrieved, by default 3
    memory_index : Optional[IndexConstructor], optional
        custom memory index, for details check `skllm.memory` submodule, by default None
    vectorizer : Optional[BaseVectorizer], optional
        scikit-llm vectorizer; if None, `OllamaVectorizer` is used, by default None
    metric : Optional[str], optional
        metric used for similarity search, by default "euclidean"
    """

    def __init__(
        self,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        default_label: str = "Random",
        prompt_template: Optional[str] = None,
        n_examples: int = 3,
        memory_index: Optional[IndexConstructor] = None,
        vectorizer: Optional[BaseVectorizer] = None,
        metric: Optional[str] = "euclidean",
        **kwargs,
    ):
        if vectorizer is None:
            vectorizer = OllamaVectorizer(model="custom_url::nomic-embed-text")
        super().__init__(
            model=model,
            default_label=default_label,
            prompt_template=prompt_template,
            n_examples=n_examples,
            memory_index=memory_index,
            vectorizer=vectorizer,
            metric=metric,
        )
        self.host = host
        self.options = options



from skllm.models._base.tagger import ExplainableNER
from skollama.llm.ollama.mixin import OllamaCompletionMixin


class OllamaExplainableNER(ExplainableNER, OllamaCompletionMixin):
    """Named entity recognition using Ollama API compatible models.

    Attributes
    ----------
    entities : dict
        dictionary of entities to recognize, with keys as entity names and values as descriptions
    display_predictions : bool, optional
        whether to display predictions, by default False
    sparse_output : bool, optional
        whether to generate a sparse representation of the predictions, by default True
    model : str, optional
        model to use, by default "llama3"
    host: str, optional
        Ollama host to connect to, by default "http://localhost:11434"
    options: dict, optional
        additional options to pass to the Ollama API, by default None
    """

    def __init__(
        self,
        entities: dict[str, str],
        display_predictions: bool = False,
        sparse_output: bool = True,
        model: str = "llama3",
        host: str = "http://localhost:11434",
        options: dict = None,
        num_workers: int = 1,
    ):
        self.model = model
        self.entities = entities
        self.display_predictions = display_predictions
        self.sparse_output = sparse_output
        self.host = host
        self.options = options
        self.num_workers = num_workers



from ollama import Client
from pydantic import BaseModel

from skllm.utils import retry


@retry(max_retries=3)
def get_chat_completion(
    messages: dict,
    model: str = "llama3",
    host: str = "http://localhost:11434",
    options: dict = None,
    format: BaseModel = "",
):
    """Gets a chat completion from an ollama server.

    Parameters
    ----------
    messages : dict
        input messages to use.
    model : str, optional
        model to use, must be available at the server
    host : str, optional
        host url to pass the requests to

    Returns
    -------
    completion : dict
    """
    client = Client(host=host)
    # completion = client.generate(model=model, messages=messages) # TODO proliferate options
    completion = client.chat(
        model=model,
        messages=messages,
        options=options,
        format=format,
    )  # TODO proliferate options
    return completion



from skllm.utils import retry
from skollama.llm.ollama.clients.ollama.credentials import set_credentials


@retry(max_retries=3)
def get_embedding(
    text: str,
    model: str = "llama3",
    api: str = "custom_url",
):
    """Encodes a string and return the embedding for a string.

    Parameters
    ----------
    text : str
        The string to encode.
    model : str, optional
        The model to use. Defaults to "text-embedding-ada-002".
    max_retries : int, optional
        The maximum number of retries to use. Defaults to 3.
    api: str, optional
        Must be custom_url currently. Defaults to "custom_url".

    Returns
    -------
    emb : list
        The GPT embedding for the string.
    """
    if api in ("custom_url"):
        client = set_credentials()  # TODO should pass host / url!
    # text = [str(t).replace("\n", " ") for t in text]
    embeddings = []
    emb = client.embeddings(
        model=model, prompt=text[0]
    )  # change to dont assume multiple texts
    e = emb["embedding"]
    if not isinstance(e, list):
        raise ValueError(
            f"Encountered unknown embedding format. Expected list, got {type(emb)}"
        )
    embeddings.append(e)
    return embeddings



from ollama import Client


def set_credentials(host: str = "http://localhost:11434"):
    """Set the OpenAI key and organization.

    Parameters
    ----------
    url : str
        The url for the ollama server.
    model : str
        The model to use.
    """
    client = Client(host=host)
    return client
